from pydantic import BaseModel, Field
from typing import Dict, List, Optional

from nsj_rest_lib2.compiler.edl_model.property_meta_model import PropertyMetaModel
from nsj_rest_lib2.compiler.edl_model.repository_model import RepositoryModel
from nsj_rest_lib2.compiler.edl_model.api_model import APIModel
from nsj_rest_lib2.compiler.edl_model.trait_property_meta_model import (
    TraitPropertyMetaModel,
)


class EntityModel(BaseModel):
    edl_version: Optional[str] = Field(default="1.0", description="Versão do EDL")
    escopo: str = Field(..., description="Escopo do EDL (define a aplicação).")
    description: str = Field(..., description="Descrição da entidade.")
    id: str = Field(
        ...,
        description="Identificador único da entidade, dentro de um determinado escopo (podendo ser sobreescrito para um tenant ou grupo _empresarial).",
    )
    trait_from: Optional[str] = Field(
        None, description="Identificador da entidade que a trait estende."
    )
    version: Optional[str] = Field(
        default="1.0", description="Versão da entidade (padrão: 1.0)."
    )
    abstract: Optional[bool] = Field(
        default=False,
        description="Indica se a entidade é abstrata (padrão: False). Caso positivo, não gera código, mas pode ser usada na geração de código de outras entidades.",
    )
    mixins: Optional[List[str]] = Field(
        None,
        description="Lista de mixins (blocos reutilizáveis) aplicados ao modelo.",
    )
    required: Optional[List[str]] = Field(
        None, description="Lista de campos obrigatórios no modelo."
    )
    partition_data: Optional[List[str]] = Field(
        None,
        description="Lista de propriedades da entidade, que serã usadas para particionamento dos dados no banco (sendo obrigatórias em todas as chamadas, inclusive como um tipo de filtro obrigatório).",
    )
    main_properties: Optional[List[str]] = Field(
        None,
        description="Lista de propriedades resumo do modelo (retornadas em qualquer chamada GET, mesmo que não pedidas por meio do parâmetro 'fields').",
    )
    search_properties: Optional[List[str]] = Field(
        None,
        description="Lista de propriedades do modelo que serão utilizadas nas chamadas de query do usuário (suportando também campos das entidades relacionadas). Sugere-se que a biblioteca crie automaticamente os índices para fulltext search.",
    )
    metric_label: Optional[List[str]] = Field(
        None,
        description="Lista de campos incluídos no agrupamento das métricas enviadas ao OpenTelemetry.",
    )
    trait_properties: Optional[Dict[str, TraitPropertyMetaModel]] = Field(
        None,
        description="Dicionário de propriedades condicionais da trait (isso é, especifica propriedades, e seus valores fixos, que serão usados como um tipo de filtro fixo usado como condição para que a propridade principal assum o papel daquele tipo de trait).",
    )
    properties: Dict[str, PropertyMetaModel] = Field(
        ..., description="Dicionário de propriedades do modelo."
    )
    repository: RepositoryModel = Field(
        ..., description="Configurações de mapeamento para o banco de dados."
    )
    api: APIModel = Field(
        ...,
        description="Definição da API REST associada ao modelo, com todos os seus endpoints.",
    )



import os
import redis

from typing import Any

REDIS_HOST = os.environ["REDIS_HOST"]
REDIS_PORT = int(os.environ["REDIS_PORT"])
REDIS_DB = int(os.getenv("REDIS_DB", 0))

redis_client = redis.Redis(host=REDIS_HOST, port=REDIS_PORT, db=REDIS_DB)


def k(*parts: str) -> str:
    return ":".join(parts)


def get_redis(*args: str) -> Any:
    value = redis_client.get(k(*args))
    if value:
        return value.decode("utf-8")
    return None


def set_redis(*args) -> None:
    value = args[-1]
    redis_client.set(k(*args[:-1]), value)


if __name__ == "__main__":
    set_redis("ping", "pong")
    print(get_redis(("ping")))



import ast

from nsj_rest_lib2.compiler.compiler_structures import (
    IndexCompilerStructure,
    PropertiesCompilerStructure,
)
from nsj_rest_lib2.compiler.edl_model.entity_model import EntityModel
from nsj_rest_lib2.compiler.edl_model.primitives import PrimitiveTypes
from nsj_rest_lib2.compiler.edl_model.property_meta_model import PropertyMetaModel
from nsj_rest_lib2.compiler.util.str_util import CompilerStrUtil
from nsj_rest_lib2.compiler.util.type_util import TypeUtil

# TODO pattern
# TODO lowercase
# TODO uppercase
# TODO Adicionar o nome da entidade, no nome das classes de enum (para evitar conflitos no caso das traits)


class EDLPropertyCompiler:
    def compile(
        self,
        properties_structure: PropertiesCompilerStructure,
        map_unique_by_property: dict[str, IndexCompilerStructure],
        entity_model: EntityModel,
    ) -> tuple[list[ast.stmt], list[ast.stmt], list[str], list[ast.stmt]]:

        # TODO Criar opção de campo calculado?

        ast_dto_attributes = []
        ast_entity_attributes = []
        props_pk = []
        enum_classes = []

        if properties_structure.properties is None:
            return (ast_dto_attributes, ast_entity_attributes, props_pk, enum_classes)

        for pkey in properties_structure.properties:
            prop = properties_structure.properties[pkey]
            enum_class_name = None

            # DTO
            ## Tratando propriedade simples (não array, não object)
            if prop.type not in [PrimitiveTypes.ARRAY, PrimitiveTypes.OBJECT]:
                keywords = []

                if prop.pk:
                    keywords.append(ast.keyword(arg="pk", value=ast.Constant(True)))
                    props_pk.append(pkey)

                if prop.key_alternative:
                    keywords.append(
                        ast.keyword(arg="candidate_key", value=ast.Constant(True))
                    )

                if (
                    properties_structure.main_properties
                    and pkey in properties_structure.main_properties
                ):
                    keywords.append(ast.keyword(arg="resume", value=ast.Constant(True)))

                if (
                    properties_structure.required
                    and pkey in properties_structure.required
                ):
                    keywords.append(
                        ast.keyword(arg="not_null", value=ast.Constant(True))
                    )

                if (
                    properties_structure.partition_data
                    and pkey in properties_structure.partition_data
                ):
                    keywords.append(
                        ast.keyword(arg="partition_data", value=ast.Constant(True))
                    )

                if pkey in map_unique_by_property:
                    unique = map_unique_by_property[pkey].index_model
                    keywords.append(
                        ast.keyword(
                            arg="unique",
                            value=ast.Constant(unique.name),
                        )
                    )

                if (
                    prop.default
                ):  # TODO Verificar esse modo de tratar valores default (principalmente expressões)
                    keywords.append(
                        ast.keyword(
                            arg="default_value",
                            value=ast.Name(str(prop.default), ctx=ast.Load()),
                        )
                    )

                if prop.trim:
                    keywords.append(ast.keyword(arg="strip", value=ast.Constant(True)))

                max = None
                min = None
                if prop.type in [PrimitiveTypes.STRING, PrimitiveTypes.EMAIL]:
                    if prop.max_length:
                        max = prop.max_length
                    elif prop.min_length:
                        min = prop.min_length
                elif prop.type in [PrimitiveTypes.INTEGER, PrimitiveTypes.NUMBER]:
                    if prop.minimum:
                        min = prop.minimum
                    elif prop.maximum:
                        max = prop.maximum

                if max:
                    keywords.append(
                        ast.keyword(arg="max", value=ast.Constant(prop.max_length))
                    )
                if min:
                    keywords.append(
                        ast.keyword(arg="min", value=ast.Constant(prop.min_length))
                    )

                if (
                    properties_structure.search_properties
                    and pkey in properties_structure.search_properties
                ):
                    keywords.append(ast.keyword(arg="search", value=ast.Constant(True)))
                else:
                    keywords.append(
                        ast.keyword(arg="search", value=ast.Constant(False))
                    )

                if (
                    properties_structure.metric_label
                    and pkey in properties_structure.metric_label
                ):
                    keywords.append(
                        ast.keyword(arg="metric_label", value=ast.Constant(True))
                    )

                if prop.type == PrimitiveTypes.CPF and not prop.validator:
                    keywords.append(
                        ast.keyword(
                            arg="validator",
                            value=ast.Attribute(
                                value=ast.Call(
                                    func=ast.Name(
                                        id="DTOFieldValidators", ctx=ast.Load()
                                    ),
                                    args=[],
                                    keywords=[],
                                ),
                                attr="validate_cpf",
                                ctx=ast.Load(),
                            ),
                        )
                    )
                elif prop.type == PrimitiveTypes.CNPJ and not prop.validator:
                    keywords.append(
                        ast.keyword(
                            arg="validator",
                            value=ast.Attribute(
                                value=ast.Call(
                                    func=ast.Name(
                                        id="DTOFieldValidators", ctx=ast.Load()
                                    ),
                                    args=[],
                                    keywords=[],
                                ),
                                attr="validate_cnpj",
                                ctx=ast.Load(),
                            ),
                        )
                    )
                elif prop.type == PrimitiveTypes.CPF_CNPJ and not prop.validator:
                    keywords.append(
                        ast.keyword(
                            arg="validator",
                            value=ast.Attribute(
                                value=ast.Call(
                                    func=ast.Name(
                                        id="DTOFieldValidators", ctx=ast.Load()
                                    ),
                                    args=[],
                                    keywords=[],
                                ),
                                attr="validate_cpf_or_cnpj",
                                ctx=ast.Load(),
                            ),
                        )
                    )
                elif prop.type == PrimitiveTypes.EMAIL and not prop.validator:
                    keywords.append(
                        ast.keyword(
                            arg="validator",
                            value=ast.Attribute(
                                value=ast.Call(
                                    func=ast.Name(
                                        id="DTOFieldValidators", ctx=ast.Load()
                                    ),
                                    args=[],
                                    keywords=[],
                                ),
                                attr="validate_email",
                                ctx=ast.Load(),
                            ),
                        )
                    )
                elif prop.validator:
                    keywords.append(
                        ast.keyword(
                            arg="validator",
                            value=ast.Name(prop.validator, ctx=ast.Load()),
                        )
                    )

                if prop.immutable:
                    keywords.append(
                        ast.keyword(arg="read_only", value=ast.Constant(True))
                    )

                if prop.on_save:
                    keywords.append(
                        ast.keyword(
                            arg="convert_to_entity",
                            value=ast.Name(prop.on_save, ctx=ast.Load()),
                        )
                    )

                if prop.on_retrieve:
                    keywords.append(
                        ast.keyword(
                            arg="convert_from_entity",
                            value=ast.Name(id=prop.on_retrieve, ctx=ast.Load()),
                        )
                    )

                if prop.domain_config:
                    result = self._compile_domain_config(pkey, prop, entity_model)
                    if not result:
                        raise Exception(
                            f"Erro desconhecido ao compilar a propriedade {pkey}"
                        )

                    enum_class_name, ast_enum_class = result
                    enum_classes.append(ast_enum_class)

                # Resolvendo o nome da propriedade no Entity
                if (
                    properties_structure.entity_properties
                    and pkey in properties_structure.entity_properties
                ):
                    entity_field_name = properties_structure.entity_properties[
                        pkey
                    ].column
                else:
                    entity_field_name = pkey

                # Escrevendo, se necessário, o alias para o nome da entity
                if entity_field_name != pkey:
                    keywords.append(
                        ast.keyword(
                            arg="entity_field",
                            value=ast.Constant(value=entity_field_name),
                        )
                    )

                # Instanciando o atributo AST
                if enum_class_name:
                    prop_type = enum_class_name
                else:
                    prop_type = TypeUtil.property_type_to_python_type(prop.type)

                ast_attr = ast.AnnAssign(
                    target=ast.Name(
                        id=CompilerStrUtil.to_snake_case(pkey), ctx=ast.Store()
                    ),
                    annotation=ast.Name(
                        id=prop_type,
                        ctx=ast.Load(),
                    ),
                    value=ast.Call(
                        func=ast.Name(id="DTOField", ctx=ast.Load()),
                        args=[],
                        keywords=keywords,
                    ),
                    simple=1,
                )

                ast_dto_attributes.append(ast_attr)

                # Entity
                ast_entity_attr = ast.AnnAssign(
                    target=ast.Name(
                        id=CompilerStrUtil.to_snake_case(entity_field_name),
                        ctx=ast.Store(),
                    ),
                    annotation=ast.Name(
                        id=TypeUtil.property_type_to_python_type(prop.type),
                        ctx=ast.Load(),
                    ),
                    value=ast.Constant(value=None),
                    simple=1,
                )

                ast_entity_attributes.append(ast_entity_attr)

        return ast_dto_attributes, ast_entity_attributes, props_pk, enum_classes

    def _compile_domain_config(
        self,
        pkey: str,
        prop: PropertyMetaModel,
        entity_model: EntityModel,
    ) -> tuple[str, ast.stmt] | None:
        if not prop.domain_config:
            return None

        # Verificando se deveria usar o mapped_value
        use_mapped_value = False
        for value in prop.domain_config:
            if value.mapped_value:
                use_mapped_value = True
                break

        # Compilando as opções do enum
        ast_values = []
        for value in prop.domain_config:
            value_name = CompilerStrUtil.to_snake_case(value.value).upper()

            if use_mapped_value and value.mapped_value is None:
                raise Exception(
                    f"Propriedade '{pkey}' possui domain_config com value '{value.value}' mas sem mapped_value"
                )

            if value.mapped_value is not None:
                ast_value = ast.Assign(
                    targets=[ast.Name(id=value_name, ctx=ast.Store())],
                    value=ast.Tuple(
                        elts=[
                            ast.Constant(value=value.value),
                            ast.Constant(value=value.mapped_value),
                        ],
                        ctx=ast.Load(),
                    ),
                )
            else:
                ast_value = ast.Assign(
                    targets=[ast.Name(id=value_name, ctx=ast.Store())],
                    value=ast.Constant(value=value.value),
                )

            ast_values.append(ast_value)

        # Instanciando o atributo AST
        enum_class_name = f"{CompilerStrUtil.to_pascal_case(entity_model.escopo)}{CompilerStrUtil.to_pascal_case(entity_model.id)}{CompilerStrUtil.to_pascal_case(pkey)}Enum"
        ast_enum_class = ast.ClassDef(
            name=enum_class_name,
            bases=[
                ast.Attribute(
                    value=ast.Name(id="enum", ctx=ast.Load()),
                    attr="Enum",
                    ctx=ast.Load(),
                )
            ],
            keywords=[],
            decorator_list=[],
            body=ast_values,
        )

        return enum_class_name, ast_enum_class



from pydantic import BaseModel, Field
from typing import List, Optional


class IndexModel(BaseModel):
    name: str = Field(..., description="Nome do índice no banco de dados.")
    columns: List[str] = Field(
        ..., description="Lista de colunas que compõem o índice."
    )
    unique: Optional[bool] = Field(
        default=False,
        description="Indica se é um índice de unicidade (padrão: False). Se não for de unicidade, será um índice usado para otimização de queries.",
    )



from typing import Any

from nsj_rest_lib2.compiler.compiler_structures import (
    IndexCompilerStructure,
    PropertiesCompilerStructure,
)
from nsj_rest_lib2.compiler.dto_compiler import DTOCompiler
from nsj_rest_lib2.compiler.edl_model.repository_model import RepositoryModel
from nsj_rest_lib2.compiler.entity_compiler import EntityCompiler
from nsj_rest_lib2.compiler.property_compiler import EDLPropertyCompiler

from nsj_rest_lib2.compiler.edl_model.entity_model import EntityModel

from nsj_rest_lib2.settings import get_logger

# TODO Atualizar o status da entidade pelo worker de compilação (e talvez parar uma compilação, quando se delete uma entidade)
# TODO Relacionamentos
# TODO Classes Abstratas
# TODO Partial Classes
# TODO Migrations


class CompilerResult:
    def __init__(self):
        self.dto_class_name: str | None = None
        self.dto_code: str | None = None
        self.entity_class_name: str | None = None
        self.entity_code: str | None = None
        self.api_expose: bool | None = None
        self.api_resource: str | None = None
        self.api_verbs: list[str] | None = None


class EDLCompiler:
    def __init__(self) -> None:
        self._properties_compiler = EDLPropertyCompiler()
        self._dto_compiler = DTOCompiler()
        self._entity_compiler = EntityCompiler()

    def compile_models(
        self, entity_models: dict[str, EntityModel]
    ) -> list[CompilerResult]:

        compiler_results = []
        for entity_model_id in entity_models:
            entity_model = entity_models[entity_model_id]
            if not entity_model.abstract:
                compiler_result = self._compile_model(entity_model, entity_models)
                compiler_results.append(compiler_result)

        return compiler_results

    def compile_model_from_edl(
        self,
        edl_json: dict[str, Any],
        dependencies_edls: list[dict[str, Any]],
    ) -> CompilerResult:
        entity_model = EntityModel(**edl_json)

        entity_models = []
        for dependency_edl in dependencies_edls:
            dependency_entity_model = EntityModel(**dependency_edl)
            entity_models.append(dependency_entity_model)

        return self.compile_model(entity_model, entity_models)

    def compile_model(
        self,
        entity_model: EntityModel,
        dependencies_models: list[EntityModel],
    ) -> CompilerResult:
        entity_models = {}
        for dependency_entity_model in dependencies_models:
            complete_entity_id = (
                f"{dependency_entity_model.escopo}/{dependency_entity_model.id}"
            )
            entity_models[complete_entity_id] = dependency_entity_model

        return self._compile_model(entity_model, entity_models)

    def _compile_model(
        self,
        entity_model: EntityModel,
        entity_models: dict[str, EntityModel],
    ) -> CompilerResult:

        # Criando um mapa de índices por nome de property
        # TODO Implementar tratamento dos índices de apoio às query (não de unicidade)
        map_indexes_by_property: dict[str, list[IndexCompilerStructure]] = {}
        map_unique_by_property: dict[str, IndexCompilerStructure] = {}
        self._make_unique_map_by_property(
            map_indexes_by_property, map_unique_by_property, entity_model, entity_models
        )

        # Criando uma cópia das coleções necessárias à compilação das properties
        # (a ideia é ser possível alterar as coleções sem afetar a entidade modelo,
        # o que será necessário para o tratamento de traits, etc - os quais serão
        # uma classe nova, resultado da união dessas propriedades).
        properties_structure = PropertiesCompilerStructure()
        self._make_properties_structures(
            properties_structure, entity_model, entity_models
        )

        # Criando a lista de atributos do DTO e da Entity; e recuperando as chaves primarias
        ast_dto_attributes, ast_entity_attributes, props_pk, enum_classes = (
            self._properties_compiler.compile(
                properties_structure,
                map_unique_by_property,
                entity_model,
            )
        )

        # Gerando o código do DTO
        dto_class_name, code_dto = self._dto_compiler.compile(
            entity_model, ast_dto_attributes, enum_classes
        )

        # Gerando o código da Entity
        entity_class_name, code_entity = self._entity_compiler.compile(
            entity_model, ast_entity_attributes, props_pk
        )

        # Construindo o resultado
        compiler_result = CompilerResult()
        compiler_result.entity_class_name = entity_class_name
        compiler_result.entity_code = code_entity
        compiler_result.dto_class_name = dto_class_name
        compiler_result.dto_code = code_dto

        # Compilando questões das APIs
        compiler_result.api_expose = entity_model.api.expose
        compiler_result.api_resource = entity_model.api.resource
        compiler_result.api_verbs = entity_model.api.verbs

        return compiler_result

    def _make_properties_structures(
        self,
        properties_structure: PropertiesCompilerStructure,
        entity_model: EntityModel,
        entity_models: dict[str, EntityModel],
    ):
        if not entity_model:
            return

        # Populando com as propriedades do trait
        if entity_model.trait_from:
            trait_model = entity_models[entity_model.trait_from]

            self._make_properties_structures(
                properties_structure,
                trait_model,
                entity_models,
            )

        # Populando com as propriedades da entidade atual
        properties_structure.properties.update(entity_model.properties)
        if entity_model.main_properties:
            properties_structure.main_properties.extend(entity_model.main_properties)
        if entity_model.required:
            properties_structure.required.extend(entity_model.required)
        if entity_model.partition_data:
            properties_structure.partition_data.extend(entity_model.partition_data)
        if entity_model.search_properties:
            properties_structure.search_properties.extend(
                entity_model.search_properties
            )
        if entity_model.metric_label:
            properties_structure.metric_label.extend(entity_model.metric_label)

        if entity_model.trait_properties:
            properties_structure.trait_properties.update(entity_model.trait_properties)

        if entity_model.repository.properties:
            properties_structure.entity_properties.update(
                entity_model.repository.properties
            )

    def _make_unique_map_by_property(
        self,
        map_indexes_by_property: dict[str, list[IndexCompilerStructure]],
        map_unique_by_property: dict[str, IndexCompilerStructure],
        entity_model: EntityModel,
        entity_models: dict[str, EntityModel],
        deep: int = 1,
    ):

        if not entity_model:
            return

        # Varrendo e organizando os índices
        if entity_model.repository.indexes:
            for index in entity_model.repository.indexes:
                for pkey in index.columns:
                    if index.unique:
                        if pkey in map_unique_by_property:
                            if deep > 1:
                                get_logger().warning(
                                    f"Propriedade '{pkey}' possui mais de um índice de unicidade (sendo um herdado). Por isso a replicação (herdada) será ignorada."
                                )
                                continue
                            else:
                                raise Exception(
                                    f"Propriedade '{pkey}' possui mais de um índice de unicidade."
                                )  # TODO Verificar esse modo de tratar erros

                        map_unique_by_property[pkey] = IndexCompilerStructure(
                            index, deep > 1
                        )
                    else:
                        list_index = map_indexes_by_property.setdefault(pkey, [])
                        list_index.append(IndexCompilerStructure(index, deep > 1))

        # Populando com as propriedades do trait
        if entity_model.trait_from:
            trait_model = entity_models[entity_model.trait_from]

            self._make_unique_map_by_property(
                map_indexes_by_property,
                map_unique_by_property,
                trait_model,
                entity_models,
                deep=deep + 1,
            )

    def list_dependencies(
        self, edl_json: dict[str, Any]
    ) -> tuple[list[str], EntityModel]:
        entity_model = EntityModel(**edl_json)

        return (self._list_dependencies(entity_model), entity_model)

    def _list_dependencies(self, entity_model: EntityModel) -> list[str]:
        entities: list[str] = []
        if entity_model.trait_from:
            entities.append(entity_model.trait_from)

        return entities


def get_files_from_directory(directory):
    files = []
    for file in os.listdir(directory):
        if file.endswith(".json"):
            files.append(os.path.join(directory, file))
    return files


if __name__ == "__main__":
    import argparse
    import json
    import os

    parser = argparse.ArgumentParser(
        description="Compila arquivos EDL para classes Python"
    )
    parser.add_argument(
        "-d",
        "--directory",
        help="Diretório com arquivos .json para compilar",
        type=str,
        required=True,
    )
    args = parser.parse_args()

    files = get_files_from_directory(args.directory)

    entities = {}
    for file in files:
        with open(file) as f:
            edl_json = json.load(f)

        # Instanciando o objeto de modelo de entidade a partir do JSON,
        # e já realizando as validações básicas de tipo e estrutura.
        print(f"Validando arquivo: {file}")
        entity_model = EntityModel(**edl_json)

        complete_entity_id = f"{entity_model.escopo}/{entity_model.id}"
        entities[complete_entity_id] = entity_model

    compiler = EDLCompiler()
    compiler_results = compiler.compile_models(entities)

    for compiler_result in compiler_results:
        print("==========================================================")
        print(f"Entity: {compiler_result.entity_class_name}")
        print(f"{compiler_result.entity_code}")
        print("\n")
        print("==========================================================")
        print(f"DTO: {compiler_result.dto_class_name}")
        print(f"{compiler_result.dto_code}")
        print("\n")

        print("==========================================================")
        print("API Expose: ", compiler_result.api_expose)
        print("API Route Path: ", compiler_result.api_resource)
        print("API Verbs: ", compiler_result.api_verbs)
        print("==========================================================")
        print("\n")



import datetime
import json
import sys
import threading
import types

from nsj_rest_lib.settings import get_logger

from nsj_rest_lib2.exception import MissingEntityConfigException
from nsj_rest_lib2.redis_config import get_redis
from nsj_rest_lib2.settings import ESCOPO_RESTLIB2, MIN_TIME_SOURCE_REFRESH


# TODO Verificar se está atualizando o loaded_at, quando o hash não muda
class LoadedEntity:
    def __init__(self):
        self.dto_class_name: str = ""
        self.entity_class_name: str = ""
        self.entity_hash: str = ""
        self.loaded_at: datetime.datetime = datetime.datetime.now()
        self.api_expose: bool = False
        self.api_verbs: list[str] = []


class Namespace:
    def __init__(self):
        self.key: str = ""
        self.loaded_entities: dict[str, LoadedEntity] = {}
        self.entities_dict: dict = None
        self.module: types.ModuleType = None


namespaces_dict: dict[str, Namespace] = {}


class EntityLoader:
    def __init__(self) -> types.NoneType:
        self._lock = threading.Lock()

    def load_entity_source(
        self,
        entity_resource: str,
        tenant: str | None,
        grupo_empresarial: str | None,
    ) -> tuple[str, str, dict, bool, list[str]]:
        # Montando as chaves dos namespaces
        grupo_key = f"tenant_{tenant}.ge_{grupo_empresarial}"
        tenant_key = f"tenant_{tenant}"
        default_key = "default"

        result = self._load_entity_source_from_memory(
            entity_resource, grupo_key, tenant_key, default_key
        )

        # Se conseguiu carregar da memória, verifica se houve alteração no hash, em relação ao redis
        if result is not None:
            # Desempacotando o result e recuperando informações do namespace
            (
                entity_config_key,
                namespace,
            ) = result

            loaded_entity = namespace.loaded_entities[entity_resource]
            dto_class_name = loaded_entity.dto_class_name
            entity_class_name = loaded_entity.entity_class_name
            entities_dict = namespace.entities_dict
            api_expose = loaded_entity.api_expose
            api_verbs = loaded_entity.api_verbs

            # Se o tempo entre o carregamento e agora for maior do que MIN_TIME_SOURCE_REFRESH minutos,
            # verifica se precisa de refresh
            time_diff = datetime.datetime.now() - loaded_entity.loaded_at

            if time_diff.total_seconds() >= MIN_TIME_SOURCE_REFRESH * 60:
                # Renovando o tempo de refresh
                loaded_entity.loaded_at = datetime.datetime.now()

                # Recuperando do Redis direto pela key (faz uma só chamada ao redis)
                loaded_config = self._load_entity_config_from_redis(
                    entity_resource,
                    grupo_key,
                    tenant_key,
                    default_key,
                    entity_config_key,
                )

                # Se não achar no redis, usa o que estava em memória
                if not loaded_config:
                    return (
                        dto_class_name,
                        entity_class_name,
                        entities_dict,
                        api_expose,
                        api_verbs,
                    )

                # Desempacotando resultado
                entity_config_key, entity_config_str = loaded_config

                # Executando o código da entidade, só se houver mudança no hash
                result_execute = self._execute_entity_source(
                    entity_config_str,
                    entity_config_key,
                    entity_resource,
                    check_refresh=True,
                )

                # Se não carregou novo código, usa o que estava em memória
                if result_execute is None:
                    return (
                        dto_class_name,
                        entity_class_name,
                        entities_dict,
                        api_expose,
                        api_verbs,
                    )
                else:
                    (
                        dto_class_name,
                        entity_class_name,
                        namespace,
                        api_expose,
                        api_verbs,
                    ) = result_execute
                    return (
                        dto_class_name,
                        entity_class_name,
                        namespace.entities_dict,
                        api_expose,
                        api_verbs,
                    )
            else:
                # Se não deu o intervalo de verificação do refresh, retorna o que está em memória
                return (
                    dto_class_name,
                    entity_class_name,
                    entities_dict,
                    api_expose,
                    api_verbs,
                )

        # Se não conseguir recuperar a entidade, procura no redis:
        loaded_config = self._load_entity_config_from_redis(
            entity_resource, grupo_key, tenant_key, default_key, None
        )

        # Se também não achar no redis, lanca exceção
        if not loaded_config:
            raise MissingEntityConfigException()

        # Desempacotando resultado
        entity_config_key, entity_config_str = loaded_config

        # Executando o código da entidade
        result_execute = self._execute_entity_source(
            entity_config_str, entity_config_key, entity_resource
        )

        if result_execute is None:
            raise RuntimeError(
                f"Erro desconhecido carregando entidade: {entity_resource}"
            )
        dto_class_name, entity_class_name, namespace, api_expose, api_verbs = (
            result_execute
        )

        return (
            dto_class_name,
            entity_class_name,
            namespace.entities_dict,
            api_expose,
            api_verbs,
        )

    def clear_namespaces(self):
        """
        Clears all loaded namespaces from memory.

        This method removes all entries from the namespaces_dict, effectively resetting
        the in-memory cache of loaded entities and their associated namespaces.
        """
        with self._lock:
            namespaces_dict.clear()

    def _execute_entity_source(
        self,
        entity_config_str: str,
        entity_config_key: str,
        entity_resource: str,
        check_refresh: bool = False,
    ) -> tuple[str, str, Namespace, bool, list[str]] | None:
        # Interpretando o json de configuração da entidade
        try:
            entity_config = json.loads(entity_config_str)

            dto_class_name = entity_config["dto_class_name"]
            entity_class_name = entity_config["entity_class_name"]
            source_dto = entity_config["source_dto"]
            source_entity = entity_config["source_entity"]
            entity_hash = entity_config["entity_hash"]

            api_expose = entity_config["api_expose"]
            # api_resource = entity_config["api_resource"]
            api_verbs = entity_config["api_verbs"]
        except json.JSONDecodeError as e:
            if not check_refresh:
                raise RuntimeError(
                    f"Erro ao decodificar JSON da entidade {entity_resource}; na chave {entity_config_key}: {e}"
                )
            else:
                get_logger().error(
                    f"Erro ao decodificar JSON da entidade {entity_resource}; na chave {entity_config_key}: {e}"
                )
                return None

        # Verificando se a entidade precisa ou não de refresh
        if check_refresh:
            loaded_namespace = namespaces_dict.get(entity_config_key)
            if not loaded_namespace:
                return None

            loaded_entity = loaded_namespace.loaded_entities.get(entity_resource)
            if not loaded_entity:
                return None

            if loaded_entity.entity_hash == entity_hash:
                return None

        # Imprimindo alerta de load no log
        get_logger().debug(
            f"Carregando entidade {entity_resource} no namespace {entity_config_key}."
        )

        # Carregando a entidade no namespace
        with self._lock:
            namespace = namespaces_dict.get(entity_config_key)
            if namespace is None:
                namespace = Namespace()
                namespace.key = entity_config_key
                namespaces_dict[entity_config_key] = namespace

            # Executando o código da entidade
            module = types.ModuleType(entity_config_key)
            sys.modules[entity_config_key] = module
            module.__dict__["__builtins__"] = __builtins__

            namespace.module = module
            namespace.entities_dict = module.__dict__

            self._safe_exec(source_dto, namespace.entities_dict, "DTO source")
            self._safe_exec(source_entity, namespace.entities_dict, "Entity source")

            # Gravando a entidade no dict de entidades carregadas
            loaded_entity = LoadedEntity()
            loaded_entity.dto_class_name = dto_class_name
            loaded_entity.entity_class_name = entity_class_name
            loaded_entity.entity_hash = entity_hash
            loaded_entity.api_expose = api_expose
            loaded_entity.api_verbs = api_verbs

            namespace.loaded_entities[entity_resource] = loaded_entity

        return (dto_class_name, entity_class_name, namespace, api_expose, api_verbs)

    def _safe_exec(self, source_code, context, description):
        try:
            exec(source_code, context)
        except Exception as e:
            get_logger().error(f"Error executing {description}: {e}")
            raise

    def _load_entity_config_from_redis(
        self,
        entity_resource: str,
        grupo_key: str,
        tenant_key: str,
        default_key: str,
        entity_config_key: str | None,
    ) -> tuple[str, str] | None:
        get_logger().debug(
            f"Procurando a configuração da entidade {entity_resource} no redis"
        )

        if entity_config_key is not None:
            entity_config_str = get_redis(
                "entity_config", ESCOPO_RESTLIB2, entity_config_key, entity_resource
            )

        else:
            entity_config_key = grupo_key
            entity_config_str = get_redis(
                "entity_config", ESCOPO_RESTLIB2, grupo_key, entity_resource
            )
            if entity_config_str is None:
                entity_config_key = tenant_key
                entity_config_str = get_redis(
                    "entity_config", ESCOPO_RESTLIB2, tenant_key, entity_resource
                )
            if entity_config_str is None:
                entity_config_key = default_key
                entity_config_str = get_redis(
                    "entity_config", ESCOPO_RESTLIB2, default_key, entity_resource
                )

        # Se não encontrar no redis, retorna None
        if entity_config_str is None:
            return None

        return (entity_config_key, entity_config_str)

    def _load_entity_source_from_memory(
        self,
        entity_resource: str,
        grupo_key: str,
        tenant_key: str,
        default_key: str,
    ) -> tuple[str, Namespace] | None:
        namespace = None
        entity_config_key = None

        # Pesquisando a entidade no namespace mais específico (grupo_empresarial)
        grupo_namespace = namespaces_dict.get(grupo_key)
        if grupo_namespace and entity_resource in grupo_namespace.loaded_entities:
            entity_config_key = grupo_key
            namespace = grupo_namespace

        # Pesquisando a entidade no namespace intermediário (tenant)
        tenant_namespace = namespaces_dict.get(tenant_key)
        if tenant_namespace and entity_resource in tenant_namespace.loaded_entities:
            entity_config_key = tenant_key
            namespace = tenant_namespace

        # Pesquisando a entidade no namespace padrão (default)
        default_namespace = namespaces_dict.get(default_key)
        if default_namespace and entity_resource in default_namespace.loaded_entities:
            entity_config_key = default_key
            namespace = default_namespace

        if namespace and entity_config_key:
            return (entity_config_key, namespace)
        else:
            return None



import logging
import os

ESCOPO_RESTLIB2 = os.environ["ESCOPO_RESTLIB2"]
APP_NAME = os.getenv("APP_NAME", "nsj_rest_lib2")

ENV = os.getenv("ENV", "dev").lower()

if ENV in ("dev", "local"):
    MIN_TIME_SOURCE_REFRESH = 0
else:
    MIN_TIME_SOURCE_REFRESH = int(os.getenv("MIN_TIME_SOURCE_REFRESH", "15"))


def get_logger():
    return logging.getLogger(APP_NAME)



from nsj_rest_lib2.compiler.edl_model.column_meta_model import ColumnMetaModel
from nsj_rest_lib2.compiler.edl_model.index_model import IndexModel
from nsj_rest_lib2.compiler.edl_model.property_meta_model import PropertyMetaModel
from nsj_rest_lib2.compiler.edl_model.trait_property_meta_model import (
    TraitPropertyMetaModel,
)


class IndexCompilerStructure:
    def __init__(self, index_model: IndexModel, inherited: bool) -> None:
        self.index_model: IndexModel = index_model
        self.inherited: bool = inherited


class PropertiesCompilerStructure:
    def __init__(self) -> None:
        self.properties: dict[str, PropertyMetaModel] = {}
        self.main_properties: list[str] = []
        self.required: list[str] = []
        self.partition_data: list[str] = []
        self.search_properties: list[str] = []
        self.metric_label: list[str] = []
        self.entity_properties: dict[str, ColumnMetaModel] = {}
        self.trait_properties: dict[str, TraitPropertyMetaModel] = {}



import enum
from typing import List


class PrimitiveTypes(enum.Enum):
    # TODO Validar esses tipos
    STRING = "string"
    NUMBER = "number"
    INTEGER = "integer"
    BOOLEAN = "boolean"
    ARRAY = "array"
    OBJECT = "object"
    UUID = "uuid"
    CURRENCY = "currency"
    QUANTITY = "quantity"
    CPF = "cpf"
    CNPJ = "cnpj"
    CPF_CNPJ = "cpf_cnpj"
    EMAIL = "email"
    DATE = "date"
    DATETIME = "datetime"


MAPPING_PRIMITIVE_TYPES_TO_PYTHON = {
    PrimitiveTypes.STRING: "str",
    PrimitiveTypes.NUMBER: "float",
    PrimitiveTypes.INTEGER: "int",
    PrimitiveTypes.BOOLEAN: "bool",
    PrimitiveTypes.ARRAY: "List",
    PrimitiveTypes.OBJECT: "dict",
    PrimitiveTypes.UUID: "uuid.UUID",
    PrimitiveTypes.CURRENCY: "float",
    PrimitiveTypes.QUANTITY: "float",
    PrimitiveTypes.CPF: "str",
    PrimitiveTypes.CNPJ: "str",
    PrimitiveTypes.CPF_CNPJ: "str",
    PrimitiveTypes.EMAIL: "str",
    PrimitiveTypes.DATE: "datetime.date",
    PrimitiveTypes.DATETIME: "datetime.datetime",
}

BasicTypes = int | bool | float | str
DefaultTypes = BasicTypes | List[BasicTypes]



import json

from typing import Any

from flask import Flask, request

from nsj_rest_lib.settings import APP_NAME

from nsj_gcf_utils.rest_error_util import format_json_error

from nsj_multi_database_lib.decorator.multi_database import multi_database

from nsj_rest_lib.controller.controller_util import DEFAULT_RESP_HEADERS
from nsj_rest_lib.controller.list_route import ListRoute
from nsj_rest_lib.controller.get_route import GetRoute
from nsj_rest_lib.controller.post_route import PostRoute
from nsj_rest_lib.controller.put_route import PutRoute
from nsj_rest_lib.controller.patch_route import PatchRoute
from nsj_rest_lib.controller.delete_route import DeleteRoute

from nsj_rest_lib2.exception import MissingEntityConfigException
from nsj_rest_lib2.service.entity_loader import EntityLoader


def _get_tenant_grupo() -> tuple[str, str]:
    # Tentando ler do query args
    query_args = request.args
    tenant = query_args.get("tenant")
    grupo_empresarial = query_args.get("grupo_empresarial")

    # Tentando ler do corpo da requisição
    try:
        body_str = request.data.decode("utf-8")
        body_json = json.loads(body_str)

        if not tenant:
            tenant = body_json.get("tenant")
        if not grupo_empresarial:
            grupo_empresarial = body_json.get("grupo_empresarial")
    except:
        pass

    return (str(tenant), str(grupo_empresarial))


def _endpoint_name(func: Any, multidb: bool, root: str) -> str:
    suffix = "_mb" if multidb else ""
    return f"{root}_{func.__name__}{suffix}"


def setup_dynamic_routes(
    flask_app: Flask,
    multidb: bool = True,
    dynamic_root_path: str = "edl1",
    injector_factory: Any = None,
) -> None:

    COLLECTION_DYNAMIC_ROUTE = f"/{APP_NAME}/{dynamic_root_path}/<entity_resource>"
    ONE_DYNAMIC_ROUTE = f"/{APP_NAME}/{dynamic_root_path}/<entity_resource>/<id>"

    def list_dynamic_wrapper(injector_factory: Any, *args: Any, **kwargs: Any) -> Any:

        def list_dynamic(*args: Any, **kwargs: Any):
            # Recuperando o identificador da entidade
            if "entity_resource" not in kwargs:
                msg = "Faltando parâmetro identificador da entidade na URL."
                return (format_json_error(msg), 400, {**DEFAULT_RESP_HEADERS})
            entity_resource = kwargs.pop("entity_resource")

            # Lendo tenant e grupo_empresarial
            tenant, grupo_empresarial = _get_tenant_grupo()

            try:
                # Recuperando o código do DTO e Entity correspondente
                entity_loader = EntityLoader()
                (
                    dto_class_name,
                    entity_class_name,
                    etities_dict,
                    api_expose,
                    api_verbs,
                ) = entity_loader.load_entity_source(
                    entity_resource, tenant, grupo_empresarial
                )

                # Verificando se essa API deve ser exposta
                if not api_expose or "GET" not in api_verbs:
                    return ("", 405, {})

                # Executando o list pelo RestLib
                route = ListRoute(
                    url=COLLECTION_DYNAMIC_ROUTE,
                    http_method="GET",
                    dto_class=etities_dict[dto_class_name],
                    entity_class=etities_dict[entity_class_name],
                    injector_factory=injector_factory,
                )

                return route.handle_request(*args, **kwargs)
            except MissingEntityConfigException:
                msg = f"Entity configuration for {entity_resource} not found."
                return (format_json_error(msg), 412, {**DEFAULT_RESP_HEADERS})

        return list_dynamic

    def get_dynamic_wrapper(injector_factory: Any, *args: Any, **kwargs: Any) -> Any:

        def get_dynamic(*args: Any, **kwargs: Any):
            # Recuperando o identificador da entidade
            if "entity_resource" not in kwargs:
                msg = "Faltando parâmetro identificador da entidade na URL."
                return (format_json_error(msg), 400, {**DEFAULT_RESP_HEADERS})
            entity_resource = kwargs.pop("entity_resource")

            # Lendo tenant e grupo_empresarial
            tenant, grupo_empresarial = _get_tenant_grupo()

            try:
                # Recuperando o código do DTO e Entity correspondente
                entity_loader = EntityLoader()
                (
                    dto_class_name,
                    entity_class_name,
                    etities_dict,
                    api_expose,
                    api_verbs,
                ) = entity_loader.load_entity_source(
                    entity_resource, tenant, grupo_empresarial
                )

                # Verificando se essa API deve ser exposta
                if not api_expose or "GET" not in api_verbs:
                    return ("", 405, {})

                # Executando o list pelo RestLib
                route = GetRoute(
                    url=ONE_DYNAMIC_ROUTE,
                    http_method="GET",
                    dto_class=etities_dict[dto_class_name],
                    entity_class=etities_dict[entity_class_name],
                    injector_factory=injector_factory,
                )

                return route.handle_request(*args, **kwargs)
            except MissingEntityConfigException:
                msg = f"Entity configuration for {entity_resource} not found."
                return (format_json_error(msg), 412, {**DEFAULT_RESP_HEADERS})

        return get_dynamic

    def post_dynamic_wrapper(injector_factory: Any, *args: Any, **kwargs: Any) -> Any:
        def post_dynamic(*args: Any, **kwargs: Any):
            # Recuperando o identificador da entidade
            if "entity_resource" not in kwargs:
                msg = "Faltando parâmetro identificador da entidade na URL."
                return (format_json_error(msg), 400, {**DEFAULT_RESP_HEADERS})
            entity_resource = kwargs.pop("entity_resource")

            # Lendo tenant e grupo_empresarial
            tenant, grupo_empresarial = _get_tenant_grupo()

            try:
                # Recuperando o código do DTO e Entity correspondente
                entity_loader = EntityLoader()
                (
                    dto_class_name,
                    entity_class_name,
                    etities_dict,
                    api_expose,
                    api_verbs,
                ) = entity_loader.load_entity_source(
                    entity_resource, tenant, grupo_empresarial
                )

                # Verificando se essa API deve ser exposta
                if not api_expose or "POST" not in api_verbs:
                    return ("", 405, {})

                # Executando o list pelo RestLib
                route = PostRoute(
                    url=COLLECTION_DYNAMIC_ROUTE,
                    http_method="POST",
                    dto_class=etities_dict[dto_class_name],
                    entity_class=etities_dict[entity_class_name],
                    injector_factory=injector_factory,
                )

                return route.handle_request(*args, **kwargs)
            except MissingEntityConfigException:
                msg = f"Entity configuration for {entity_resource} not found."
                return (format_json_error(msg), 412, {**DEFAULT_RESP_HEADERS})

        return post_dynamic

    def put_dynamic_wrapper(injector_factory: Any, *args: Any, **kwargs: Any) -> Any:
        def put_dynamic(*args: Any, **kwargs: Any):
            # Recuperando o identificador da entidade
            if "entity_resource" not in kwargs:
                msg = "Faltando parâmetro identificador da entidade na URL."
                return (format_json_error(msg), 400, {**DEFAULT_RESP_HEADERS})
            entity_resource = kwargs.pop("entity_resource")

            # Lendo tenant e grupo_empresarial
            tenant, grupo_empresarial = _get_tenant_grupo()

            try:
                # Recuperando o código do DTO e Entity correspondente
                entity_loader = EntityLoader()
                (
                    dto_class_name,
                    entity_class_name,
                    etities_dict,
                    api_expose,
                    api_verbs,
                ) = entity_loader.load_entity_source(
                    entity_resource, tenant, grupo_empresarial
                )

                # Verificando se essa API deve ser exposta
                if not api_expose or "PUT" not in api_verbs:
                    return ("", 405, {})

                # Executando o list pelo RestLib
                route = PutRoute(
                    url=ONE_DYNAMIC_ROUTE,
                    http_method="PUT",
                    dto_class=etities_dict[dto_class_name],
                    entity_class=etities_dict[entity_class_name],
                    injector_factory=injector_factory,
                )

                return route.handle_request(*args, **kwargs)
            except MissingEntityConfigException:
                msg = f"Entity configuration for {entity_resource} not found."
                return (format_json_error(msg), 412, {**DEFAULT_RESP_HEADERS})

        return put_dynamic

    def patch_dynamic_wrapper(injector_factory: Any, *args: Any, **kwargs: Any) -> Any:
        def patch_dynamic(*args: Any, **kwargs: Any):
            # Recuperando o identificador da entidade
            if "entity_resource" not in kwargs:
                msg = "Faltando parâmetro identificador da entidade na URL."
                return (format_json_error(msg), 400, {**DEFAULT_RESP_HEADERS})
            entity_resource = kwargs.pop("entity_resource")

            # Lendo tenant e grupo_empresarial
            tenant, grupo_empresarial = _get_tenant_grupo()

            try:
                # Recuperando o código do DTO e Entity correspondente
                entity_loader = EntityLoader()
                (
                    dto_class_name,
                    entity_class_name,
                    etities_dict,
                    api_expose,
                    api_verbs,
                ) = entity_loader.load_entity_source(
                    entity_resource, tenant, grupo_empresarial
                )

                # Verificando se essa API deve ser exposta
                if not api_expose or "PATCH" not in api_verbs:
                    return ("", 405, {})

                # Executando o list pelo RestLib
                route = PatchRoute(
                    url=ONE_DYNAMIC_ROUTE,
                    http_method="PATCH",
                    dto_class=etities_dict[dto_class_name],
                    entity_class=etities_dict[entity_class_name],
                    injector_factory=injector_factory,
                )

                return route.handle_request(*args, **kwargs)
            except MissingEntityConfigException:
                msg = f"Entity configuration for {entity_resource} not found."
                return (format_json_error(msg), 412, {**DEFAULT_RESP_HEADERS})

        return patch_dynamic

    def delete_dynamic_wrapper(injector_factory: Any, *args: Any, **kwargs: Any) -> Any:
        def delete_dynamic(*args: Any, **kwargs: Any):
            # Recuperando o identificador da entidade
            if "entity_resource" not in kwargs:
                msg = "Faltando parâmetro identificador da entidade na URL."
                return (format_json_error(msg), 400, {**DEFAULT_RESP_HEADERS})
            entity_resource = kwargs.pop("entity_resource")

            # Lendo tenant e grupo_empresarial
            tenant, grupo_empresarial = _get_tenant_grupo()

            try:
                # Recuperando o código do DTO e Entity correspondente
                entity_loader = EntityLoader()
                (
                    dto_class_name,
                    entity_class_name,
                    etities_dict,
                    api_expose,
                    api_verbs,
                ) = entity_loader.load_entity_source(
                    entity_resource, tenant, grupo_empresarial
                )

                # Verificando se essa API deve ser exposta
                if not api_expose or "DELETE" not in api_verbs:
                    return ("", 405, {})

                # Executando o list pelo RestLib
                route = DeleteRoute(
                    url=ONE_DYNAMIC_ROUTE,
                    http_method="DELETE",
                    dto_class=etities_dict[dto_class_name],
                    entity_class=etities_dict[entity_class_name],
                    injector_factory=injector_factory,
                )

                return route.handle_request(*args, **kwargs)
            except MissingEntityConfigException:
                msg = f"Entity configuration for {entity_resource} not found."
                return (format_json_error(msg), 412, {**DEFAULT_RESP_HEADERS})

        return delete_dynamic

    # Ajustando para o padrão com multi database (se necessário)
    if multidb:
        list_dynamic = multi_database()(list_dynamic_wrapper(injector_factory))
        get_dynamic = multi_database()(get_dynamic_wrapper(injector_factory))
        post_dynamic = multi_database()(post_dynamic_wrapper(injector_factory))
        put_dynamic = multi_database()(put_dynamic_wrapper(injector_factory))
        patch_dynamic = multi_database()(patch_dynamic_wrapper(injector_factory))
        delete_dynamic = multi_database()(delete_dynamic_wrapper(injector_factory))
    else:
        list_dynamic = list_dynamic_wrapper(injector_factory)
        get_dynamic = get_dynamic_wrapper(injector_factory)
        post_dynamic = post_dynamic_wrapper(injector_factory)
        put_dynamic = put_dynamic_wrapper(injector_factory)
        patch_dynamic = patch_dynamic_wrapper(injector_factory)
        delete_dynamic = delete_dynamic_wrapper(injector_factory)

    # Registrando as rotas no flask
    flask_app.add_url_rule(
        COLLECTION_DYNAMIC_ROUTE,
        endpoint=_endpoint_name(list_dynamic, multidb, dynamic_root_path),
        view_func=list_dynamic,
        methods=["GET"],
    )
    flask_app.add_url_rule(
        ONE_DYNAMIC_ROUTE,
        endpoint=_endpoint_name(get_dynamic, multidb, dynamic_root_path),
        view_func=get_dynamic,
        methods=["GET"],
    )
    flask_app.add_url_rule(
        COLLECTION_DYNAMIC_ROUTE,
        endpoint=_endpoint_name(post_dynamic, multidb, dynamic_root_path),
        view_func=post_dynamic,
        methods=["POST"],
    )
    flask_app.add_url_rule(
        ONE_DYNAMIC_ROUTE,
        endpoint=_endpoint_name(put_dynamic, multidb, dynamic_root_path),
        view_func=put_dynamic,
        methods=["PUT"],
    )
    flask_app.add_url_rule(
        ONE_DYNAMIC_ROUTE,
        endpoint=_endpoint_name(patch_dynamic, multidb, dynamic_root_path),
        view_func=patch_dynamic,
        methods=["PATCH"],
    )
    flask_app.add_url_rule(
        ONE_DYNAMIC_ROUTE,
        endpoint=_endpoint_name(delete_dynamic, multidb, dynamic_root_path),
        view_func=delete_dynamic,
        methods=["DELETE"],
    )
